from enum import Enum


class PropertyType(str, Enum):
    """
    Enum of all supported property types.
    """
    BOOL = 'bool'
    INT = 'int'
    FLOAT = 'float'
    DOUBLE = 'double'
    STRING = 'string'
    REGEX = 'regex'



from .name_converter import NamingConventionType
from .generator_naming_conventions import GeneratorNamingConventions


class LanguageConfigNamingConventions(GeneratorNamingConventions):
    """
    Encapsulates the naming conventions which are used by the LanguageConfigBase class.
    """
    file_naming_convention: NamingConventionType = None



from enum import IntEnum, auto
import re


class NamingConventionType(IntEnum):
    """
    Enum of all supported naming conventions.
    """
    SNAKE_CASE = 1
    SCREAMING_SNAKE_CASE = auto()
    CAMEL_CASE = auto()
    PASCAL_CASE = auto()
    KEBAP_CASE = auto()


class UnknownNamingConventionException(Exception):
    def __init__(self, naming_convention_type: NamingConventionType):
        super().__init__(f'Unknown naming convention type {naming_convention_type}')


class NameConverter:
    """
    Is used to convert a provided string into a specific string case (e.g., snake-case, camel-case, ...).
    """

    @staticmethod
    def convert(name: str, type: NamingConventionType) -> str:
        """
        Converts the provided string to the convention specified by type.

        :param name: String to convert.
        :type name:  str
        :param type: Naming convention to use.
        :type type:  NamingConventionType

        :raises UnknownNamingConventionException: Raised if an unknown naming convention type is used.

        :return: Converted string.
        :rtype:  str
        """
        UNDERLINE = '_'

        # Handle Camel- or Pascal-cased strings by replacing uppercase letters with
        # lowercase and add an underline in front.
        name = re.sub(r'[A-Z]', lambda match: f'{UNDERLINE}{match.group(0).lower()}', name).lstrip(UNDERLINE)

        # Replace all special characters by underline for easier further processing.
        name = re.sub(r'\W+', UNDERLINE, name)

        if type == NamingConventionType.SNAKE_CASE:
            name = name.lower()
        elif type == NamingConventionType.SCREAMING_SNAKE_CASE:
            name = name.upper()
        elif type == NamingConventionType.CAMEL_CASE or type == NamingConventionType.PASCAL_CASE:
            # Replace all special characters followed by a letter by the uppercase version of the letter.
            compare_name = ''
            while compare_name != name:
                compare_name = name
                name = re.sub(rf'{UNDERLINE}+([a-zA-Z0-9])',
                    lambda match: match.group(1).upper(), name # Thanks for the hint: https://stackoverflow.com/a/8934655.
                )

            # If Pascal-case, uppercase the first letter.
            if type == NamingConventionType.PASCAL_CASE:
                name = f'{name[0].upper()}{name[1:]}'
        elif type == NamingConventionType.KEBAP_CASE:
            name = re.sub(rf'{UNDERLINE}+', '-', name)
        else:
            raise UnknownNamingConventionException(type)
            
        return name



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Dict, Tuple

from .distribute_info import DistributeInfo
from .distributor_credentials import DistributorCredentials


class DistributorBase(ABC):
    """
    Abstract class that acts as the base for all Distributor implementations. Distributors
    are used to distribute the generated configs to different locations (based on the actual
    distributor implementation).
    """

    def __init__(self, config: Dict, credentials: DistributorCredentials=None):
        super().__init__()

        self._config = config
        self._credentials = credentials

    def from_config(self, key: str) -> Tuple[any, bool]:
        """
        Retrieves a value from the distributor config. If the key doesn't exist,
        None is returned.

        :param key: Value key.
        :type key:  str

        :return: Returns a tuple where the first entry is the value and the second
                 a boolean which states if the key exists.
        :rtype:  (any, bool)
        """
        key_exists = key in self._config

        return self._config[key] if key_exists else None, key_exists
    
    def distribute(self, file_name: str, data: str):
        """
        Distributes the config according to the derivative implementation.

        :param file_name: Config file name.
        :type file_name:  str
        :param data:      Config file data.
        :type data:       str

        :return: The current instance.
        :rtype:  DistributorBase
        """
        self._distribute(DistributeInfo(
            file_name,
            data,
            self._credentials,
        ))
        return self

    @abstractmethod
    def _distribute(self, info: DistributeInfo):

        """
        Method to distribute a generated config which must be implemented by a derivative class.

        :param info: Contains the required information to distribute the generated config.
        :type info:  DistributeInfo
        """
        pass



from enum import IntEnum, auto
from importlib_metadata import entry_points  # Since importlib.metadata changes way too often, use importlib_metadata.
import re
from typing import List, Type

from .distributor_base import DistributorBase
from .language_config_base import LanguageConfigBase


class PluginType(IntEnum):
    UNKNOWN = 0,
    LANGUAGE_CONFIG = auto()
    DISTRIBUTOR = auto()


class Plugin:
    def __init__(self, name: str, class_type: Type) -> None:
        self._name = name
        self._class_type = class_type

        plugin_type = PluginType.UNKNOWN

        # Specify plugin type by base-class.
        if self._is_language_plugin(class_type):
            plugin_type = PluginType.LANGUAGE_CONFIG
        elif self._is_distributor_plugin(class_type):
            plugin_type = PluginType.DISTRIBUTOR
        self._type = plugin_type

    def get_name(self) -> str:
        return self._name
    
    def get_type(self) -> PluginType:
        return self._type
    
    def get_class_type(self) -> Type:
        return self._class_type
    
    def _inherits(self, check_type: Type, check_class: Type):
        base_classes_names = list(map(lambda clazz: clazz.__name__, check_class.__bases__))
        return check_type.__name__ in base_classes_names
    
    def _is_language_plugin(self, check_class: Type):
        return self._inherits(LanguageConfigBase, check_class)
    
    def _is_distributor_plugin(self, check_class: Type):
        return self._inherits(DistributorBase, check_class)


class PluginManager:
    def __init__(self, plugins: List[Plugin]=None) -> None:
        self._plugins: List[Plugin] = []

        # Since a default list cannot be assigned to parameters in the method header, because it only gets initialized
        # once and then the list gets re-used (see https://stackoverflow.com/a/1145781), make sure to set undefined
        # variables to list (see also https://docs.python.org/3/reference/compound_stmts.html#function-definitions).
        if not plugins:
            plugins = []

        self._load_plugins()
        self.add_plugins(plugins)

    def add_plugins(self, plugins: List[Plugin], replace=True):
        def plugin_name(plugin: Plugin):
            return plugin.get_name().replace('-', '_').strip()

        # Added plugins overwrite loaded ones if names match, doubles are removed.
        for plugin in [p for p in plugins if p and p.get_type() != PluginType.UNKNOWN]:
            found = False

            if replace:
                for i, plugin_temp in enumerate(self._plugins):
                    if plugin_name(plugin_temp) == plugin_name(plugin):

                        # If not found yet, replace the plugin, otherwise remove it.
                        if not found:
                            self._plugins[i] = plugin
                        else:
                            del self._plugins[i]
                        found = True

            # If not found, append the plugin to the list.
            if not found:
                self._plugins.append(plugin)

        return self

    def get_plugins(self) -> List[Type[Plugin]]:
        return self._plugins

    def get_language_config_plugins(self) -> List[Type[Plugin]]:
        return self._get_plugins_by_type(PluginType.LANGUAGE_CONFIG)

    def get_distributor_plugins(self) -> List[Type[Plugin]]:
        return self._get_plugins_by_type(PluginType.DISTRIBUTOR)
    
    def _get_plugins_by_type(self, type: PluginType):
        return [plugin for plugin in self._plugins if plugin.get_type() == type]

    def _load_plugins(self):
        plugins = []

        for entry_point in [e for e in entry_points() if re.match('ninja(-|_)bear(-|_).+', e.group)]:
            plugin_class = entry_point.load()

            if plugin_class:
                plugins.append(Plugin(entry_point.group, plugin_class))
        self.add_plugins(plugins)



import argparse
from os import path
from typing import List

from .base.orchestrator import Orchestrator
from .base.distributor_credentials import DistributorCredentials

_CONFIG_PARAMETER = 'config'
_OUTPUT_PARAMETER = 'output'
_SECRET_PARAMETER = 'secret'
_DISTRIBUTE_PARAMETER = 'distribute'


def _parse_credentials(credential_strings: List[str]) -> List[DistributorCredentials]:
    credentials = []

    for credential_string in credential_strings if credential_strings else []:
        front_and_rear_parts = list(filter(lambda part: part, credential_string.split('=')))
        rear_parts = list(filter(
            lambda part: part,
            front_and_rear_parts[1].split(':') if len(front_and_rear_parts) == 2 else [],
        ))
        alias = front_and_rear_parts[0] if len(front_and_rear_parts) > 0 else None
        user = rear_parts[0] if len(rear_parts) > 1 else None
        password = rear_parts[1] if len(rear_parts) > 1 else (
            rear_parts[0] if len(rear_parts) == 1 else None
        )

        if alias and password:
            credentials.append(DistributorCredentials(alias, user, password))
    return credentials


def main():
    parser = argparse.ArgumentParser()

    parser.add_argument('-c', f'--{_CONFIG_PARAMETER}', help='Path to configuration file', required=True, type=str)
    parser.add_argument('-o', f'--{_OUTPUT_PARAMETER}', help='Output location', required=False, type=str, default='.')
    parser.add_argument('-s', f'--{_SECRET_PARAMETER}',
        help='Credential for distributions in the form of <alias>=[<username>:]<password>',
        required=False, action='append')
    parser.add_argument('-d', f'--{_DISTRIBUTE_PARAMETER}',
        help='Distribute the generated configs to the specified locations', required=False, action='store_true')

    args = parser.parse_args()

    # TODO: Might also strip backslashes.
    output_dir = f'{str(getattr(args, _OUTPUT_PARAMETER)).strip("/")}/' if hasattr(args, _OUTPUT_PARAMETER) else ''

    if output_dir and not path.isdir(output_dir):
        raise Exception(f'Output directory {output_dir} does not exist')
    
    credentials = _parse_credentials(getattr(args, _SECRET_PARAMETER) if hasattr(args, _SECRET_PARAMETER) else [])
    config = Orchestrator.read_config(getattr(args, _CONFIG_PARAMETER), credentials)
    config.write(output_dir)

    if getattr(args, _DISTRIBUTE_PARAMETER):
        config.distribute()


if __name__ == '__main__':
    main()



from os import path
import os
import pathlib
import shutil
from typing import Dict, List, Type
import unittest

from src.ninja_bear import GeneratorBase, PropertyType, NamingConventionType, DumpInfo, DistributeInfo, Plugin
from src.ninja_bear.base.orchestrator import Orchestrator
from src.ninja_bear.base.generator_configuration import GeneratorConfiguration
from src.ninja_bear.base.language_config_base import LanguageConfigBase
from src.ninja_bear.base.distributor_base import DistributorBase
from src.ninja_bear.base.distributor_credentials import DistributorCredentials


_COMPARE_FILE_CONTENT = """
struct TestConfig:
    boolean myBoolean = true
    int myInteger = 142
    float myFloat = 322.0
    float myCombinedFloat = 45724.0
    double myDouble = 233.9
    regex myRegex = /Test Reg(E|e)x/ -- Just another RegEx.
    string mySubstitutedString = 'Sometimes I just want to scream Hello Mars!'
    string myCombinedString = 'I am telling you that this string got included from test-include.yaml.'
"""


class ExampleScriptGenerator(GeneratorBase):
    """
    ExampleScript specific generator. For more information about the generator methods, refer to GeneratorBase.
    """

    def _default_type_naming_convention(self) -> NamingConventionType:
        return NamingConventionType.PASCAL_CASE
    
    def _line_comment(self, string: str) -> str:
        return f'-- {string}'
    
    def _dump(self, info: DumpInfo) -> str:
        code = f'struct {info.type_name}:\n'

        for property in info.properties:
            type = property.type
            value = property.value

            if type == PropertyType.BOOL:
                type_string = 'boolean'
                value = 'true' if value else 'false'
            elif type == PropertyType.INT:
                type_string = 'int'
            elif type == PropertyType.FLOAT:
                type_string = 'float'
            elif type == PropertyType.DOUBLE:
                type_string = 'double'
            elif type == PropertyType.STRING:
                type_string = 'string'
                value = f'\'{value}\''
            elif type == PropertyType.REGEX:
                type_string = 'regex'
                value = f'/{value}/'

            comment = f' {self._line_comment(property.comment)}' if property.comment else ''
            code += f'{" " * info.indent}{type_string} {property.name} = {value}{comment}\n'

        return code


class ExampleScriptConfig(LanguageConfigBase):
    """
    ExampleScript specific config. For more information about the config methods, refer to LanguageConfigBase.
    """

    def _file_extension(self) -> str:
        return 'es'

    def _generator_type(self) -> Type[ExampleScriptGenerator]:
        return ExampleScriptGenerator
    
    def _default_file_naming_convention(self) -> NamingConventionType:
        return NamingConventionType.KEBAP_CASE

    def _allowed_file_name_pattern(self) -> str:
        return r'.+'
    

class ExampleDistributor(DistributorBase):
    def __init__(self, config: Dict, credentials: DistributorCredentials=None) -> DistributorBase:
        super().__init__(config, credentials)

    def _distribute(self, info: DistributeInfo):
        return self


class Test(unittest.TestCase):
    def __init__(self, methodName: str = "runTest") -> None:
        super().__init__(methodName)
        self._test_path = pathlib.Path(__file__).parent.resolve()
        self._test_config_path = path.join(self._test_path, '..', 'example/test-config.yaml')
        self._plugins = [
            Plugin('examplescript', ExampleScriptConfig),
            Plugin('exampledistributor', ExampleDistributor),
        ]

    def test_read_config(self):
        orchestrator = Orchestrator.read_config(self._test_config_path, plugins=self._plugins)
        self._evaluate_configs(orchestrator.language_configs)

    def test_parse_config(self):
        TEST_INCLUDE = 'test-include.yaml'

        with open(self._test_config_path, 'r') as f:
            content = f.read().replace(TEST_INCLUDE, os.path.join(os.getcwd(), 'example', TEST_INCLUDE))
        orchestrator = Orchestrator.parse_config(content, 'test-config', plugins=self._plugins)

        self._evaluate_configs(orchestrator.language_configs)

    def test_run_generators(self):
        orchestrator = Orchestrator.read_config(self._test_config_path, plugins=self._plugins)
        language_configs = orchestrator.language_configs

        self.assertEqual(len(language_configs), 1)

        language_config = language_configs[0]
        config_generator = language_config.generator
        local_generator = ExampleScriptGenerator(
            GeneratorConfiguration(
                indent=config_generator._indent,
                transformers=config_generator.transformers,
                naming_conventions=config_generator._naming_conventions,
                type_name=config_generator._type_name
            ),
            properties=config_generator._properties,
            additional_props=config_generator._additional_props,
        )

        original_max_diff = self.maxDiff
        self.maxDiff = None

        self.assertEqual(local_generator.dump().strip(), _COMPARE_FILE_CONTENT.strip())
        self.maxDiff = original_max_diff

    def test_write_configs(self):
        OUTPUT_DIR = path.join(self._test_path, 'test_output')
        orchestrator = Orchestrator.read_config(self._test_config_path, plugins=self._plugins)

        if not os.path.isdir(OUTPUT_DIR):
            os.mkdir(OUTPUT_DIR)
        
        # Write all configs to the output folder.
        orchestrator.write(OUTPUT_DIR)

        # Collect the output file names.
        files = os.listdir(OUTPUT_DIR)

        # Cleanup output directory.
        shutil.rmtree(OUTPUT_DIR)

        # Compare files.
        for config in orchestrator.language_configs:
            self.assertIn(config.config_info.file_name_full, files)

    def test_distribution(self):
        # Get secret from environment variables.
        credential = DistributorCredentials('example-alias', None, 'password')
        orchestrator = Orchestrator.read_config(self._test_config_path, [credential], plugins=self._plugins)

        self._evaluate_configs(orchestrator.language_configs)
        orchestrator.distribute()

    def _evaluate_configs(self, configs: List[LanguageConfigBase]):
        self.assertEqual(len(configs), 1)
        self._evaluate_config(configs[0])

    def _evaluate_config(self, config: LanguageConfigBase):
        self.assertIsNotNone(config)
        self.assertEqual(config.config_info.file_extension, 'es')
        self.assertEqual(config.config_info.file_name, 'TestConfig')

    def _evaluate_common_properties(
        self,
        config: LanguageConfigBase,
        extension: str,
        name: str,
    ):
        self.assertEqual(config.config_info.file_extension, extension)
        self.assertEqual(config.config_info.file_name, name)



class NoAliasProvidedException(Exception):
    def __init__(self):
        super().__init__('No alias has been provided')


class DistributorCredentials:
    """
    Class to encapsulate credentials for specific distributor types.
    """

    distributor_alias: str
    user: str
    password: str

    def __init__(self, distributor_alias: str, user: str='', password: str=''):
        """
        DistributorBase constructor.

        :param distributor_alias: Alias to identify the credentials.
        :type distributor_alias:  str
        :param user:              Credential user, defaults to ''
        :type user:               str, optional
        :param password:          Credential password, defaults to ''
        :type password:           str, optional

        :raises NoAliasProvidedException: Raised if no distribution alias has been provided.
        """
        # Make sure there's an alias for the credentials.
        if not distributor_alias:
            raise NoAliasProvidedException()

        self.distributor_alias = distributor_alias
        self.user = user
        self.password = password



from dataclasses import dataclass
from ..base.generator_naming_conventions import GeneratorNamingConventions

from .configuration_base import ConfigurationBase


@dataclass(kw_only=True)  # https://medium.com/@aniscampos/python-dataclass-inheritance-finally-686eaf60fbb5
class GeneratorConfiguration(ConfigurationBase):
    """
    Encapsulates the configuration properties used by the GeneratorBase class.
    """
    type_name: str
    """
    Name of the generated type. HINT: This acts more like a template than the
    real name as some conventions must be met and therefore the default convention
    specified by the deriving class will be used if no naming convention for the
    type name was provided (see _default_type_naming_convention).
    """
    naming_conventions: GeneratorNamingConventions = None
    """
    Specifies which case convention to use for the properties. If not provided,
    the name as specified will be used. Defaults to None
    """



from __future__ import annotations
import os
import re
from typing import Dict, List, Tuple, Type

import yaml
from schema import Schema, Use, Optional, Or

from .plugin_manager import Plugin, PluginManager, PluginType
from .name_converter import NamingConventionType
from .property import Property
from .property_type import PropertyType
from .language_config_base import LanguageConfigBase
from .language_config_naming_conventions import LanguageConfigNamingConventions
from .distributor_base import DistributorBase
from .distributor_credentials import DistributorCredentials
from .meta_data_settings import MetaDataSettings

# Main keys.
_KEY_INCLUDES = 'includes'
_KEY_TRANSFORMERS = 'transformers'
_KEY_DISTRIBUTORS = 'distributors'
_KEY_LANGUAGES = 'languages'
_KEY_PROPERTIES = 'properties'
_KEY_META = 'meta'

# Common keys.
_KEY_AS = 'as'
_KEY_IGNORE = 'ignore'

# Include keys.
_INCLUDE_KEY_PATH = 'path'

# Transformer keys.
_TRANSFORMER_KEY_TRANSFORMER = 'transformer'

# Distributor keys.
_DISTRIBUTOR_KEY_DISTRIBUTOR = 'distributor'

# Language keys.
_LANGUAGE_KEY_LANGUAGE = 'language'
_LANGUAGE_KEY_FILE_NAMING = 'file_naming'
_LANGUAGE_KEY_PROPERTY_NAMING = 'property_naming'
_LANGUAGE_KEY_TYPE_NAMING = 'type_naming'
_LANGUAGE_KEY_INDENT = 'indent'
_LANGUAGE_KEY_TRANSFORMERS = _KEY_TRANSFORMERS
_LANGUAGE_KEY_DISTRIBUTORS = _KEY_DISTRIBUTORS

_LANGUAGE_KEY_TYPE = 'type'
_LANGUAGE_KEY_NAME = 'name'

# Property keys.
_PROPERTY_KEY_VALUE = 'value'
_PROPERTY_KEY_HIDDEN = 'hidden'
_PROPERTY_KEY_COMMENT = 'comment'

# Meta keys.
_META_KEY_USER = 'user'
_META_KEY_DATE = 'date'
_META_KEY_TIME = 'time'
_META_KEY_VERSION = 'version'
_META_KEY_LINK = 'link'


class UnknownPropertyTypeException(Exception):
    def __init__(self, property_type: str):
        super().__init__(f'Unknown property type {property_type}')


class SeveralPluginsException(Exception):
    def __init__(self, type: str, name: str):
        super().__init__(f'Several {type} plugins found for {name}')


class SeveralLanguagePluginsException(SeveralPluginsException):
    def __init__(self, language: str):
        super().__init__('language', language)


class NoLanguagePluginException(Exception):
    def __init__(self, language_names: List[str]):
        super().__init__(f'No language plugin found for {" or ".join(language_names)}')


class AliasAlreadyInUseException(Exception):
    def __init__(self, alias: str):
        super().__init__(f'The include-alias \'{alias}\' is already in use')


class SeveralDistributorPluginsException(SeveralPluginsException):
    def __init__(self, distributor: str):
        super().__init__('distributor', distributor)


class DistributorNotFoundException(Exception):
    def __init__(self, distributor: str):
        super().__init__(f'The distributor \'{distributor}\' could not be found')


class DefinitionAliasNotFoundException(Exception):
    def __init__(self, type: str, alias: str):
        super().__init__(f'The {type} alias \'{alias}\' could not be found')


class TransformerAliasNotFoundException(DefinitionAliasNotFoundException):
    def __init__(self, alias: str):
        super().__init__('transformer', alias)


class DistributorAliasNotFoundException(DefinitionAliasNotFoundException):
    def __init__(self, alias: str):
        super().__init__('distributor', alias)


class Config:
    """
    Handles the config evaluation by parsing the provided YAML string via the parse-method.
    """

    @staticmethod
    def read(
        path: str,
        distributor_credentials: List[DistributorCredentials]=None,
        plugins: List[Plugin]=None,
    ) -> List[LanguageConfigBase]:
        """
        Reads the provided YAML configuration file and generates a list of language configurations.

        :param path:                    Path to load the YAML file from (see example/test-config.yaml for configuration
                                        details).
        :type path:                     str
        :param distributor_credentials: Credentials for distributors, defaults to None
        :type distributor_credentials:  List[DistributorCredentials], optional
        :param plugins:                 Caller-provided plugins (overwrite loaded plugins), defaults to None
        :type plugins:                  List[Plugin], optional

        :return: Language configurations which further can be dumped as config files.
        :rtype:  List[LanguageConfigBase]
        """
        return Config._read(
            path,
            distributor_credentials=distributor_credentials,
            plugins=plugins
        )[0]

    @staticmethod
    def parse(
        content: str | object,
        config_name: str,
        distributor_credentials: List[DistributorCredentials]=None,
        plugins: List[Plugin]=None,
    ) -> List[LanguageConfigBase]:
        """
        Parses the provided YAML configuration string and returns the corresponding language configurations.

        :param content:                 YAML configuration string. For config details, please check the test-config.yaml
                                        in the example folder. HINT: This can also be an already parsed object.
        :type content:                  str | object
        :param config_name:             Output config file name. NOTE: The actual file name format might be overruled by
                                        the specified file_naming rule from the config.
        :type config_name:              str
        :param distributor_credentials: Credentials for distributors, defaults to None
        :type distributor_credentials:  List[DistributorCredentials], optional
        :param plugins:                 Caller-provided plugins (overwrite loaded plugins), defaults to None
        :type plugins:                  List[Plugin], optional

        :return: Language configurations which further can be dumped as config files.
        :rtype:  List[LanguageConfigBase]
        """
        return Config._parse(
            content,
            config_name,
            distributor_credentials=distributor_credentials,
            plugins=plugins
        )[0]

    @staticmethod
    def _read(
        path: str,
        namespace: str='',
        namespaces: List[str]=None,
        distributor_credentials: List[DistributorCredentials]=None,
        plugins: List[Plugin]=None,
    ) -> List[LanguageConfigBase]:
        """
        Reads the provided YAML configuration file and generates a list of language configurations.

        :param path:                    Path to load the YAML file from (see example/test-config.yaml for configuration
                                        details).
        :type path:                     str
        :param namespace:               Specifies a namespace for the config. If None or empty, no namespace will
                                        be set.
        :type nammespace:               str
        :param namespaces:              List of namespaces.
        :type nammespaces:              List[str]
        :param distributor_credentials: Credentials for distributors, defaults to None
        :type distributor_credentials:  List[DistributorCredentials], optional
        :param plugins:                 Caller-provided plugins (overwrite loaded plugins), defaults to None
        :type plugins:                  List[Plugin], optional

        :return: Language configurations which further can be dumped as config files.
        :rtype:  List[LanguageConfigBase]
        """
        with open(path, 'r') as f:
            content = f.read()

        # Prepare config name.
        last_part = path.replace(r'\\', '/').split('/')[-1]

        if '.' in last_part:
            config_name = '.'.join(last_part.split('.')[0:-1])
        else:
            config_name = last_part
        return Config._parse(
            content,
            config_name,
            namespace,
            os.path.dirname(path),
            namespaces,
            distributor_credentials,
            plugins,
        )

    @staticmethod
    def _parse(
        content: str | object,
        config_name: str,
        namespace: str='',
        directory: str='',
        namespaces: List[str]=None,
        distributor_credentials: List[DistributorCredentials]=None,
        plugins: List[Plugin]=None,
    ) -> Tuple[List[LanguageConfigBase], List[Property]]:
        """
        Parses the provided YAML configuration string and returns the corresponding language configurations.

        :param content:                 YAML configuration strings. For config details, please check the
                                        test-config.yaml in the example folder.
        :type content:                  str
        :param config_name:             Output config file name. NOTE: The actual file name format might be overruled by
                                        the specified file_naming rule from the config.
        :type config_name:              str
        :param namespace:               Specifies a namespace for the config. If None or empty, no namespace will
                                        be set.
        :type nammespace:               str
        :param namespaces:              List of namespaces.
        :type nammespaces:              List[str]
        :param distributor_credentials: Credentials for distributors, defaults to None
        :type distributor_credentials:  List[DistributorCredentials], optional
        :param plugins:                 Caller-provided plugins (overwrite loaded plugins), defaults to None
        :type plugins:                  List[Plugin], optional

        :raises AliasAlreadyInUseException: Raised if an included config file uses an already defined alias.

        :return: Language configurations which further can be dumped as config files.
        :rtype:  List[LanguageConfigBase]
        """
        plugin_manager = PluginManager(plugins)
        yaml_object = yaml.safe_load(content) if isinstance(content, str) else content
        validated_object = Config._schema().validate(yaml_object)
        language_configs: List[LanguageConfigBase] = []
        properties: List[Property] = []
        language_config_plugins = plugin_manager.get_language_config_plugins()
        transformers = Config._evaluate_transformers(validated_object)
        distributor_plugins = plugin_manager.get_distributor_plugins()
        distributors = Config._evaluate_distributors(validated_object, distributor_plugins, distributor_credentials)
        meta_data_settings = Config._evaluate_meta_data_settings(validated_object)

        # Since a default list cannot be assigned to parameters in the method header, because it only gets initialized
        # once and then the list gets re-used (see https://stackoverflow.com/a/1145781), make sure to set undefined
        # variables to list (see also https://docs.python.org/3/reference/compound_stmts.html#function-definitions).
        if not namespaces:
            namespaces = []
        if not distributor_credentials:
            distributor_credentials = []
        if not plugins:
            plugins = []

        # Evaluate included files and their properties.
        if _KEY_INCLUDES in validated_object:
            for inclusion in validated_object[_KEY_INCLUDES]:
                ignore = inclusion[_KEY_IGNORE] if _KEY_IGNORE in inclusion else False

                # If inclusion shall not be ignored, include it.
                if not ignore:
                    inclusion_namespace = inclusion[_KEY_AS]

                    # Make sure that a included config file does not re-define an alias.
                    if inclusion_namespace in namespaces:
                        raise AliasAlreadyInUseException(inclusion_namespace)
                    else:
                        namespaces.append(inclusion_namespace)
                    inclusion_path = inclusion[_INCLUDE_KEY_PATH]

                    # If the provided path is relative, incorporate the provided directory into the path.
                    if not os.path.isabs(inclusion_path):
                        inclusion_path = os.path.join(directory, inclusion_path)

                    # Read included config and put properties into property list.
                    for inclusion_property in Config._read(inclusion_path, inclusion_namespace, namespaces)[1]:
                        inclusion_property.hidden = True  # Included properties are not being exported by default.
                        properties.append(inclusion_property)
                

        # Collect properties as they are the same for all languages.
        for property in validated_object[_KEY_PROPERTIES]:
            ignore = property[_KEY_IGNORE] if _KEY_IGNORE in property else False

            # If property shall not be ignored, include it.
            if not ignore:
                properties.append(Property(
                    name=property[_LANGUAGE_KEY_NAME],
                    value=property[_PROPERTY_KEY_VALUE],
                    property_type=property[_LANGUAGE_KEY_TYPE],
                    hidden=property[_PROPERTY_KEY_HIDDEN] if _PROPERTY_KEY_HIDDEN in property else None,
                    comment=property[_PROPERTY_KEY_COMMENT] if _PROPERTY_KEY_COMMENT in property else None,
                    namespace=namespace,
                ))

        # Evaluate each language setting one by one.
        if _KEY_LANGUAGES in validated_object:
            for language in validated_object[_KEY_LANGUAGES]:
                ignore = language[_KEY_IGNORE] if _KEY_IGNORE in language else False

                # If language shall not be ignored, include it.
                if not ignore:
                    naming_conventions = LanguageConfigNamingConventions()
                    language_name = language[_LANGUAGE_KEY_LANGUAGE]
                    indent = language[_LANGUAGE_KEY_INDENT] if _LANGUAGE_KEY_INDENT in language else None

                    # Evaluate language.

                    # Evaluate file naming-convention.
                    naming_conventions.file_naming_convention = Config._evaluate_naming_convention_type(
                        language[_LANGUAGE_KEY_FILE_NAMING] if _LANGUAGE_KEY_FILE_NAMING in language else None
                    )

                    # Evaluate properties naming-convention.
                    naming_conventions.properties_naming_convention = Config._evaluate_naming_convention_type(
                        language[_LANGUAGE_KEY_PROPERTY_NAMING] if _LANGUAGE_KEY_PROPERTY_NAMING in language else None
                    )

                    # Evaluate type naming-convention.
                    naming_conventions.type_naming_convention = Config._evaluate_naming_convention_type(
                        language[_LANGUAGE_KEY_TYPE_NAMING] if _LANGUAGE_KEY_TYPE_NAMING in language else None
                    )
                    config_type = Config._evaluate_language_config(language_config_plugins, language_name)

                    language_configs.append(config_type(
                        config_name=config_name,
                        properties=properties,
                        indent=indent,
                        transformers=Config._evaluate_language_transformers(language, transformers),
                        naming_conventions=naming_conventions,
                        distributors=Config._evaluate_language_distributors(language, distributors),
                        meta_data_settings=meta_data_settings,

                        # Pass all language props as additional_props to let the specific
                        # generator decide which props it requires additionally.
                        additional_props=language,
                    ))

        return language_configs, properties
    
    @staticmethod
    def _schema() -> Schema:
        """
        Returns the config validation schema.

        :return: Config validation schema.
        :rtype:  Schema
        """
        return Schema({
            Optional(_KEY_INCLUDES): [{
                _INCLUDE_KEY_PATH: str,
                _KEY_AS: str,
                Optional(_KEY_IGNORE): bool,
            }],
            Optional(_KEY_TRANSFORMERS): [{
                _TRANSFORMER_KEY_TRANSFORMER: str,
                _KEY_AS: str,
                Optional(_KEY_IGNORE): bool,
            }],
            Optional(_KEY_DISTRIBUTORS): [{
                _DISTRIBUTOR_KEY_DISTRIBUTOR: str,
                _KEY_AS: str,
                Optional(_KEY_IGNORE): bool,
                Optional(object): object  # Collect other properties.
            }],
            Optional(_KEY_LANGUAGES): [{
                _LANGUAGE_KEY_LANGUAGE: str,
                Optional(_LANGUAGE_KEY_FILE_NAMING): str,
                Optional(_LANGUAGE_KEY_INDENT): int,
                Optional(_LANGUAGE_KEY_TRANSFORMERS): [str],
                Optional(_LANGUAGE_KEY_DISTRIBUTORS): [str],
                Optional(_KEY_IGNORE): bool,
                Optional(object): object  # Collect other properties.
            }],
            _KEY_PROPERTIES: [{
                _LANGUAGE_KEY_TYPE: Use(Config._evaluate_data_type),
                _LANGUAGE_KEY_NAME: str,
                _PROPERTY_KEY_VALUE: Or(str, bool, int, float),
                Optional(_PROPERTY_KEY_HIDDEN): bool,
                Optional(_PROPERTY_KEY_COMMENT): str,
                Optional(_KEY_IGNORE): bool,
            }],
            Optional(_KEY_META): {
                Optional(_META_KEY_USER): bool,
                Optional(_META_KEY_DATE): bool,
                Optional(_META_KEY_TIME): bool,
                Optional(_META_KEY_VERSION): bool,
                Optional(_META_KEY_LINK): bool,
            },
        })

    @staticmethod
    def _plugin_names(prefix: str, plugin_name: str) -> List[str]:
        NINJA_BEAR_PLUGIN_PREFIX = 'ninja-bear-'

        cleaned_prefix = re.sub(rf'^{NINJA_BEAR_PLUGIN_PREFIX}', '', prefix).strip('-')
        prefix = f'{NINJA_BEAR_PLUGIN_PREFIX}{cleaned_prefix}-'

        # Remove ninja-bear prefix.
        plugin_name_cleaned = re.sub(rf'^{prefix}', '', plugin_name)

        # Add prefix.
        prefixed_plugin_name_cleaned = f'{prefix}{plugin_name_cleaned}'

        def replace_dashes(s: str):
            return s.replace('-', '_')

        # Create possible plugin names.
        return list(set([
            plugin_name_cleaned,
            prefixed_plugin_name_cleaned,
            replace_dashes(plugin_name_cleaned),
            replace_dashes(prefixed_plugin_name_cleaned),
        ]))
    
    @staticmethod
    def _evaluate_meta_data_settings(validated_object: object) -> MetaDataSettings:
        """
        Evaluates the specified meta data settings.

        :param validated_object: Schema validated config object.
        :type validated_object:  object

        :return: Dictionary of defined transformer-scripts where the key is the alias.
        :rtype:  MetaDataSettings
        """
        meta_data_settings = MetaDataSettings()

        if _KEY_META in validated_object:
            settings = validated_object[_KEY_META]

            def from_settings(key: str) -> bool:
                return settings[key] if key in settings else None

            meta_data_settings.user = from_settings(_META_KEY_USER)
            meta_data_settings.date = from_settings(_META_KEY_DATE)
            meta_data_settings.time = from_settings(_META_KEY_TIME)
            meta_data_settings.version = from_settings(_META_KEY_VERSION)
            meta_data_settings.link = from_settings(_META_KEY_LINK)

        return meta_data_settings

    @staticmethod
    def _evaluate_language_config(language_plugins: List[Plugin], language_name: str) -> Type[LanguageConfigBase]:
        """
        Evaluates the corresponding language config from a language plugin list for the given language name.

        :param language_plugins: List of language plugins to search in.
        :type language_plugins:  List[Plugin]
        :param language_name:    Language name to look for.
        :type language_name:     str

        :raises SeveralLanguagePluginsException: Raised if several plugins were found for the requested language.
        :raises NoLanguagePluginException:       Raised if an unsupported language was used in the config.

        :return: The corresponding language config class.
        :rtype:  Type[LanguageConfigBase]
        """
        language_config_type = None

        # Make sure only language configs get processed.
        language_plugins = [p for p in language_plugins if p.get_type() == PluginType.LANGUAGE_CONFIG]

        # Create possible language names.
        language_names = Config._plugin_names('language', language_name)

        for plugin in language_plugins:
            if plugin.get_name() in language_names:
                if not language_config_type:
                    language_config_type = plugin.get_class_type()
                else:
                    raise SeveralLanguagePluginsException(language_name)

        if not language_config_type:
            raise NoLanguagePluginException(language_names)
        return language_config_type

    
    @staticmethod
    def _evaluate_data_type(type: str) -> PropertyType:
        """
        Evaluates a properties data type.

        :param type: Property type string (e.g., bool | string | ...).
        :type type:  str

        :raises UnknownPropertyTypeException: Raised if an unsupported property type was used in the config.

        :return: The corresponding PropertyType enum value.
        :rtype:  PropertyType
        """
        try:
            type = PropertyType(type)
        except ValueError:
            raise UnknownPropertyTypeException(type)
        return type
    
    @staticmethod
    def _evaluate_transformers(validated_object: object) -> Dict[str, DistributorBase]:
        """
        Evaluates specified transformers.

        :param validated_object: Schema validated config object.
        :type validated_object:  object

        :return: Dictionary of defined transformer-scripts where the key is the alias.
        :rtype:  Dict[str, str]
        """
        transformers = {}

        if _KEY_TRANSFORMERS in validated_object:
            transformer_configs = validated_object[_KEY_TRANSFORMERS]

            for transformer_config in transformer_configs:
                ignore = transformer_config[_KEY_IGNORE] if _KEY_IGNORE in transformer_config else False

                # If distributor shall not be ignored, include it.
                if not ignore:
                    def from_config(key: str):
                        return transformer_config[key] if key in transformer_config else None

                    transformer_script = from_config(_TRANSFORMER_KEY_TRANSFORMER)
                    alias = from_config(_KEY_AS)

                    transformers[alias] = transformer_script

        return transformers
    
    @staticmethod
    def _evaluate_language_referenced_definitions(
        language_config: Dict[str, any],
        definitions: Dict[str, any],
        key: str,
        exception_type: Type[DefinitionAliasNotFoundException],
    ) -> List[any]:
        """
        Evaluates specified transformers of a language.

        :param language_config: Language config object.
        :type language_config:  Dict[str, any]
        :param definitions:     Dictionary of defined definitions (e.g. distributors) where the key is the alias.
        :type definitions:      Dict[str, any]
        :param key:             Key to look for in the language config.
        :type key:              str
        :param exception_type:  Exception type to instantiate if an alias could not be found.
        :type exception_type:   Type[DefinitionAliasNotFoundException]

        :return: List of evaluated definitions (e.g. distributors) for the given language.
        :rtype:  List[any]
        """
        language_definitions = []

        # This time better safe than sorry.
        if not language_config:
            language_config = {}
        if not definitions:
            definitions = {}

        # Get definition references if provided.
        definition_aliases = language_config[key] if key in language_config else []
        
        # Collect all definitions that belong to the language.
        for language_alias in definition_aliases:
            if language_alias not in definitions:
                raise exception_type(language_alias)
            language_definitions.append(definitions[language_alias])

        return language_definitions
    
    @staticmethod
    def _evaluate_language_transformers(
        language_config: Dict[str, any],
        transformers: Dict[str, str]
    ) -> List[str]:
        """
        Evaluates specified transformers of a language.

        :param language_config: Language config object.
        :type language_config:  Dict[str, any]
        :param distributors:    Dictionary of defined transformer scripts where the key is the alias.
        :type distributors:     Dict[str, str], optional

        :return: List of evaluated transformer scripts for the given language.
        :rtype:  List[str]
        """
        return Config._evaluate_language_referenced_definitions(
            language_config,
            transformers,
            _LANGUAGE_KEY_TRANSFORMERS,
            TransformerAliasNotFoundException,
        )
    
    @staticmethod
    def _evaluate_distributors(
        validated_object: object,
        distributor_plugins: List[Plugin]=None,
        distributor_credentials: List[DistributorCredentials]=None
    ) -> Dict[str, DistributorBase]:
        """
        Evaluates specified distributors.

        :param validated_object:        Schema validated config object.
        :type validated_object:         object
        :param distributor_plugins:     List of all discovered distributor plugins, defaults to None
        :type distributor_plugins:      List[Plugin], optional
        :param distributor_credentials: Potentially required credentials, defaults to None
        :type distributor_credentials:  List[DistributorCredential], optional

        :return: Dictionary of defined distributors where the key is the alias.
        :rtype:  Dict[str, DistributorBase]
        """
        distributors = {}

        if _KEY_DISTRIBUTORS in validated_object:
            distributor_configs = validated_object[_KEY_DISTRIBUTORS]
            credentials_map = {}

            # Since a default list cannot be assigned to parameters in the method header, because it only gets
            # initialized once and then the list gets re-used (see https://stackoverflow.com/a/1145781), make
            # sure to set undefined variables to list (see also
            # https://docs.python.org/3/reference/compound_stmts.html#function-definitions).
            if not distributor_plugins:
                distributor_plugins = []
            if not distributor_credentials:
                distributor_credentials = []

            # Make sure only language configs get processed.
            distributor_plugins = [p for p in distributor_plugins if p.get_type() == PluginType.DISTRIBUTOR]

            # Map credential list to dictionary based on the credential alias for easer access.
            for distributor_credential in distributor_credentials:
                credentials_map[distributor_credential.distributor_alias] = distributor_credential

            for distributor_config in distributor_configs:
                ignore = distributor_config[_KEY_IGNORE] if _KEY_IGNORE in distributor_config else False

                # If distributor shall not be ignored, include it.
                if not ignore:
                    def from_config(key: str):
                        return distributor_config[key] if key in distributor_config else None

                    distributor_name = from_config(_DISTRIBUTOR_KEY_DISTRIBUTOR)

                    # Create possible distributor names.
                    distributor_names = Config._plugin_names('distributor', distributor_name)

                    found_distributors_classes = [
                        plugin.get_class_type() for plugin in distributor_plugins if
                        plugin.get_name() in distributor_names
                    ]
                    alias = from_config(_KEY_AS)
                    length = len(found_distributors_classes)

                    if length == 1:
                        found_distributor_class = found_distributors_classes[0]
                        distributors[alias] = found_distributor_class(
                            distributor_config,
                            credentials_map[alias] if alias in credentials_map else None
                        )
                    elif length > 1:
                        raise SeveralDistributorPluginsException(distributor_name)
                    else:
                        raise DistributorNotFoundException(distributor_name)

        return distributors
    
    @staticmethod
    def _evaluate_language_distributors(
        language_config: Dict[str, any],
        distributors: Dict[str, DistributorBase]
    ) -> List[DistributorBase]:
        """
        Evaluates specified distributors of a language.

        :param language_config: Language config object.
        :type language_config:  Dict[str, any]
        :param distributors:    Dictionary of defined distributors where the key is the alias.
        :type distributors:     Dict[str, DistributorBase], optional

        :return: List of evaluated distributors for the given language.
        :rtype:  List[DistributorBase]
        """
        return Config._evaluate_language_referenced_definitions(
            language_config,
            distributors,
            _LANGUAGE_KEY_DISTRIBUTORS,
            DistributorAliasNotFoundException,
        )

    @staticmethod
    def _evaluate_naming_convention_type(naming_convention: str) -> NamingConventionType:
        """
        Evaluates which naming convention type to use for the output file.

        :param naming_convention: Naming convention string (e.g., snake | camel | ...).
        :type naming_convention:  str

        :return: The corresponding NamingConventionType enum value.
        :rtype:  NamingConventionType
        """
        if naming_convention:
            if naming_convention == 'snake':
                naming_convention = NamingConventionType.SNAKE_CASE
            elif naming_convention.replace('_', '-') == 'screaming-snake':
                naming_convention = NamingConventionType.SCREAMING_SNAKE_CASE
            elif naming_convention == 'camel':
                naming_convention = NamingConventionType.CAMEL_CASE
            elif naming_convention == 'pascal':
                naming_convention = NamingConventionType.PASCAL_CASE
            elif naming_convention == 'kebap':
                naming_convention = NamingConventionType.KEBAP_CASE
        return naming_convention



from __future__ import annotations
import copy
import random
import re
from typing import Callable, List
from .property_type import PropertyType


class UnknownSubstitutionException(Exception):
    def __init__(self, substitution_property: str):
        super().__init__(f'Unknown substitution property {substitution_property}')


class RecursiveSubstitutionException(Exception):
    def __init__(self, substitution_property: str):
        super().__init__(f'It\'s not allowed for a property to reference itself ({substitution_property})')


class InvalidVariableNameException(Exception):
    def __init__(self, variable_name: str):
        super().__init__(f'{variable_name} is not a valid variable name')


class InvalidNamespaceException(Exception):
    def __init__(self, namespace: str):
        super().__init__(f'{namespace} is not a valid namespace')


class Property:
    _PROPERTY_SUBSTITUTION_PATTERN = r'\${((_|[a-zA-Z])(\w|-)*((\.(_|[a-zA-Z])(\w|-)*))?)}'
    _NAMING_PATTERN = r'^(_|[a-zA-Z])(\w|-)*$'  # Define a general naming pattern.
    _PROPERTY_NAME_PATTERN = _NAMING_PATTERN
    _NAMESPACE_NAME_PATTERN = _NAMING_PATTERN

    """
    Holds all the required information to create a property string.

    :raises InvalidVariableNameException: Raised if an invalid variable name has been provided.
    """

    def __init__(
        self,
        name: str,
        value: str | bool | int | float,
        property_type: PropertyType,
        hidden: bool = False,
        comment: str = None,
        namespace: str = None,
    ):
        """
        Constructor

        :param name:          Property name.
        :type name:           str
        :param value:         Property value.
        :type value:          str | bool | int | float
        :param property_type: Property type.
        :type property_type:  PropertyType
        :param hidden:        If True, the property will not be put out to the generated file, defaults to False
        :type hidden:         bool, optional
        :param comment:       Property description, defaults to None
        :type comment:        str, optional
        :param namespace:     Namespace to give the property an additional information to where it belongs, defaults to
                              None to avoid collision with other properties with the same name (e.g., imported from
                              other files).
        :type namespace:      str, optional

        :raises InvalidVariableNameException: Raised if an invalid variable name has been provided.
        :raises InvalidNamespaceException:    Raised if an invalid namespace has been provided.
        """
        # Check if the key is a valid variable name.
        if not re.match(Property._PROPERTY_NAME_PATTERN, name):
            raise InvalidVariableNameException(name)
        
        # Check if the key is a valid variable name.
        if namespace and not re.match(Property._NAMESPACE_NAME_PATTERN, namespace):
            raise InvalidNamespaceException(namespace)
        
        # Make sure that the provided value is valid even if it's a string.
        value = Property._convert_value(value, property_type)
        
        self.name = name
        self.type = property_type
        self.value = value
        self.hidden = hidden
        self.comment = comment
        self.namespace = namespace

    @staticmethod
    def substitute(property: Property, properties: List[Property]) -> None:
        """
        Substitutes the referenced property references.

        :param property:   Property which's value shall be updated.
        :type property:    Property
        :param properties: List of properties to get the substitution values from.
        :type properties:  List[Property]

        :raises UnknownSubstitutionException:   Raised if the requested substitution property does not exist.
        :raises RecursiveSubstitutionException: Raised if a property referenced itself as substitution.
        """
        # Copy properties to avoid messing around with the originals.
        properties_copy = copy.deepcopy(properties)

        def replace(match, _: Property):

            def add_namespace(property_name, namespace):
                return f'{namespace}.{property_name}'

            # Incorporate namespace into property name.
            def prepared_property_name(propertyTemp):
                return add_namespace(propertyTemp.name, propertyTemp.namespace) \
                    if propertyTemp.namespace else propertyTemp.name

            substitution_property_value = match.group(1)  # E.g. myReplaceString or ti.myReplaceString.
            property_name = prepared_property_name(property)
            
            if property.namespace:
                substitution_property_value = add_namespace(substitution_property_value, property.namespace)
            
            # Substitute property only if it's not the same property as the one which is currently being processed.
            if substitution_property_value != property_name:
                found_properties = [
                    search_property.value for search_property in properties_copy if
                    prepared_property_name(search_property) == substitution_property_value
                ]

                if not found_properties:
                    raise UnknownSubstitutionException(substitution_property_value)
                replacement = found_properties[0]
            else:
                # TODO: Handle indirect self reference.
                raise RecursiveSubstitutionException(
                    f'Property {property_name} must not reference itself!'
                )
            return f'{replacement}'
        
        original_value = property.value
        Property._replace_property_value(property, replace)

        # Check if value changed due to substitution.
        if property.value != original_value:
            # IF type is some kind of number or boolean, evaluate the new string and assign the result to the
            # current property.
            if property.type in [PropertyType.INT, PropertyType.FLOAT, PropertyType.DOUBLE, PropertyType.BOOL]:
                property.value = Property._convert_value(eval(property.value), property.type)

    @staticmethod
    def _convert_value(value: any, property_type: PropertyType) -> any:
        if isinstance(value, str):
            if property_type == PropertyType.BOOL:
                # Correct boolean property value to 'true' or 'false'.
                value = False if value.lower() in ['0', 'false', 'no', 'off'] else True
            elif property_type == PropertyType.INT:
                # If numbers can be substituted validly and produce another number, keep it as string.
                if not Property._is_valid_number_substitution(value):
                    match = re.match(r'\d+', value)
                    value = int(match.group(0)) if match else 0  # Remove everything that comes after the integer.
            elif property_type == PropertyType.FLOAT or property_type == PropertyType.DOUBLE:
                # If numbers can be substituted validly and produce another number, keep it as string.
                if not Property._is_valid_number_substitution(value):
                    match = re.match(r'\d+(\.\d+)?', value)
                    value = float(match.group(0)) if match else 0  # Remove everything that comes after the float.
        return value

    @staticmethod
    def _replace_property_value(property: Property, callout: Callable[[re.Match, Property], str]) -> None:
        """
        Replaces the substitution strings of a property value by using the provided callout function.

        :param property: Property of which the value shall be be processed.
        :type property:  Property
        :param callout:  Callout function to which the RegEx match is passed. It must return a replacement string.
        :type callout:   Callable[[re.Match], str]
        """
        def callout_wrapper(match: re.Match) -> str:
            return callout(match, property)

        if isinstance(property.value, str):
            # Replace all occurrences.
            property.value = re.sub(Property._PROPERTY_SUBSTITUTION_PATTERN, callout_wrapper, property.value)

    @staticmethod
    def _is_valid_number_substitution(value: any) -> bool:
        """
        Checks if the provided value can be converted to a number value if all referenced substitution values get
        replaced by random integers.

        :param value: Value to check.
        :type  value: any

        :return: True if substituted value would result in a number.
        :rtype:  bool
        """
        valid = False

        if isinstance(value, str):
            pseudo_property = Property('p', value, PropertyType.STRING)

            # Substitute with pseudo data.
            def replace(*_):
                random_number = random.randint(1, 10)
                return f'{random_number}'
            Property._replace_property_value(pseudo_property, replace)

            try:
                eval(pseudo_property.value)
                valid = True
            except Exception:
                # Nothing to do here, value was invalid.
                pass
        return valid



from .name_converter import NamingConventionType


class GeneratorNamingConventions:
    """
    Encapsulates the naming conventions which are used by the GeneratorBase class.
    """
    properties_naming_convention: NamingConventionType = None
    type_naming_convention: NamingConventionType = None



from __future__ import annotations
from abc import ABC, abstractmethod
import copy
import datetime
import getpass
from typing import Dict, List

from .info import VERSION
from .configuration_base import _DEFAULT_INDENT
from .generator_configuration import GeneratorConfiguration
from .generator_naming_conventions import GeneratorNamingConventions
from .name_converter import NamingConventionType, NameConverter
from .property import Property
from .dump_info import DumpInfo


class PropertyAlreadyExistsException(Exception):
    def __init__(self, property: str):
        super().__init__(f'Property {property} already exists')


class NoTypeNameProvidedException(Exception):
    def __init__(self):
        super().__init__('No type name has been provided')


class InvalidDumpTypeException(Exception):
    def __init__(self):
        super().__init__('The returned dump value is not a string')


class GeneratorBase(ABC):
    """
    Abstract class that acts as the base for all Generator implementations.
    """

    def __init__(
        self,
        config: GeneratorConfiguration,
        properties: List[Property]=None,
        additional_props: Dict[str, any]=None
    ):
        """
        Constructor

        :param config:           Generator configuration.
        :type config:            GeneratorConfiguration
        :param properties:       List of properties to generator by the GeneratorBase derivate, defaults to None
        :type properties:        List[Property], optional
        :param additional_props: All props that might need to be used by the derivating class, defaults to None
        :type additional_props:  Dict[str, any], optional
        """
        type_name = config.type_name
        indent = config.indent

        # Since a default list cannot be assigned to parameters in the method header, because it only gets initialized
        # once and then the list gets re-used (see https://stackoverflow.com/a/1145781), make sure to set undefined
        # variables to list (see also https://docs.python.org/3/reference/compound_stmts.html#function-definitions).
        if not properties:
            properties = []
        if not additional_props:
            additional_props = {}

        self.transformers = config.transformers
        self._meta_data_settings = config.meta_data_settings
        self._properties: List[Property] = []
        self._naming_conventions = \
            config.naming_conventions if config.naming_conventions else GeneratorNamingConventions()
        self._additional_props = additional_props

        self._set_type_name(type_name)
        self.set_indent(indent)

        # Add properties one by one.
        [self.add_property(property) for property in properties]

    def add_property(self, property: Property):
        """
        Adds a property to the properties list. IMPORTANT: Property names must be unique.

        :param property: Property to add.
        :type property:  Property

        :raises PropertyAlreadyExistsException: Raised if the instance already contains a property with the same name.

        :return: The current generator instance.
        :rtype:  Self
        """
        found_property = len([
            p for p in self._properties if p.name == property.name and p.namespace == property.namespace
        ]) > 0

        # Make sure that the name doesn't already exist.
        if found_property:
            raise PropertyAlreadyExistsException(property.name)

        self._properties.append(property)
        return self

    def set_indent(self, indent: int):
        """
        Sets the whitespace indent for the properties.

        :param indent: Indent value. If this value is less than 0, _DEFAULT_INDENT gets used.
        :type indent:  int

        :return: The current generator instance.
        :rtype:  Self
        """
        self._indent = indent if indent and indent >= 0 else _DEFAULT_INDENT
        return self

    def dump(self) -> str:
        """
        Generates a config file string.

        :return: Config file string.
        :rtype:  str
        """
        
        # Create copies of the properties to avoid messing around with the originals.
        properties_copy = [copy.deepcopy(property) for property in self._properties]

        # Transform properties if transform function was provided.
        self._apply_transformations(properties_copy)

        # Substitute property values.
        for property in properties_copy:
            Property.substitute(property, properties_copy)

        # Remove hidden properties.
        properties_copy = [property for property in properties_copy if not property.hidden]

        # If not naming conventation has been provided, use camel-case as default.
        if not self._naming_conventions.properties_naming_convention:
            self._naming_conventions.properties_naming_convention = NamingConventionType.CAMEL_CASE

        # Update property names according to naming convention.
        for property in properties_copy:
            property.name = NameConverter.convert(
                property.name, 
                self._naming_conventions.properties_naming_convention
            )

        s = self._dump(DumpInfo(
            self._type_name,
            properties_copy,
            self._indent,
            self._additional_props,
        ))

        # Make sure a string has been returned from _dump.
        if not isinstance(s, str):
            raise InvalidDumpTypeException()
        
        # Add meta data if required.
        s = self._add_meta_data(s)

        return self._add_newline(s)
    
    def get_type_name(self) -> str:
        """
        Returns the evaluated type name.

        :return: Evaluated type name.
        :rtype:  str
        """
        return self._type_name
    
    @abstractmethod
    def _default_type_naming_convention(self) -> NamingConventionType:
        """
        Abstract method which must be implemented by the deriving class to specify the default type naming convention.

        :return: Default naming convention.
        :rtype:  NamingConventionType
        """
        pass

    @abstractmethod
    def _line_comment(self, string: str) -> str:
        """
        Abstract method which must be implemented by the deriving class to turn a string into a line comment.

        :param string: String to turn into a line comment.
        :type string:  str

        :return: Commented string.
        :rtype:  str
        """
        pass

    @abstractmethod
    def _dump(self, info: DumpInfo) -> str:
        """
        Abstract method which must be implemented by the deriving class to create a type string.

        :param type_name:  Contains to required information to dump language specific code.
        :type type_name:   DumpInfo

        :return: Dumped type string.
        :rtype:  str
        """
        pass

    def _add_newline(self, s: str) -> str:
        """
        Adds a trailing newline if required.

        :param s: String to add the newline to.
        :type s:  str

        :return: Updated string.
        :rtype:  str
        """
        if s[-1] != '\n':
            s += '\n'
        return s
    
    def _add_meta_data(self, s: str) -> str:
        """
        Adds a meta data comment to the generated config string.

        :param s: Config string.
        :type s:  str

        :return: Updated config string.
        :rtype:  str
        """
        settings = self._meta_data_settings

        if settings:
            meta_data = {}

            if settings.user:
                meta_data['user'] = getpass.getuser()  # https://bugs.python.org/issue40821#msg383161
            if settings.date:
                meta_data['date'] = datetime.date.today().isoformat()
            if settings.time:
                meta_data['time'] = datetime.datetime.now().time().isoformat()
            if settings.version:
                meta_data['version'] = VERSION
            if settings.link:
                meta_data['link'] = 'https://pypi.org/project/ninja-bear/'

            items = meta_data.items()
            
            if len(items):
                s = self._add_newline(s) + '\n'

                for attribute, value in items:
                    s += self._add_newline(
                        self._line_comment(f'{attribute}: {str(value)}')
                    )
        return s

    def _set_type_name(self, name: str):
        """
        Sets the type name to the specified name. If no naming convention was set, the default
        naming convention, specified by the deriving class, will be used.

        :param name: Name of the generated type. HINT: This acts more like a template than the
                     real name as some conventions must be met and therefore the default convention
                     specified by the deriving class will be used if no naming convention for the
                     type name was provided (see _default_type_naming_convention).
        :type name:  str

        :raises NoTypeNameProvidedException: Raised if no name has been provided.

        :return: The current generator instance.
        :rtype:  Self
        """
        if not name:
            raise NoTypeNameProvidedException()
        naming_convention = self._naming_conventions.type_naming_convention

        self._type_name = NameConverter.convert(
            name,

            # Evaluate type naming convention. Use default if none was provided.
            naming_convention if naming_convention else self._default_type_naming_convention()
        )
        return self
    
    def _apply_transformations(self, properties_copy: List[Property]) -> None:
        """
        Applies the user defined value transformation to each property value.

        :param properties_copy: Copy of all properties (to prevent modification of original).
        :type properties_copy:  List[Property]
        """
        if self.transformers:
            NAME_KEY = 'name'
            VALUE_KEY = 'value'
            TYPE_KEY = 'type'
            PROPERTIES_KEY = 'properties'

            for i, property in enumerate(properties_copy):
                # Create dictionary for local variables. This dictionary will also be used
                # to get the modified value afterwards (https://stackoverflow.com/a/67824076).
                local_variables = {
                    NAME_KEY: property.name,
                    VALUE_KEY: property.value,
                    TYPE_KEY: property.type.value,
                    PROPERTIES_KEY: properties_copy,
                }

                # Execute user defined Python scripts to transform properties.
                for transformer in self.transformers:
                    exec(transformer, None, local_variables)
                    
                    # Create new property from modified value.
                    properties_copy[i] = Property(
                        name=property.name,
                        value=local_variables[VALUE_KEY],
                        property_type=property.type,
                        hidden=property.hidden,
                        comment=property.comment,
                        namespace=property.namespace,
                    )



from .base.info import VERSION  # noqa: F401
from .base.orchestrator import Orchestrator  # noqa: F401
from .base.distributor_base import DistributorBase  # noqa: F401
from .base.distributor_credentials import DistributorCredentials  # noqa: F401
from .base.language_config_base import LanguageConfigBase  # noqa: F401
from .base.generator_base import GeneratorBase  # noqa: F401
from .base.dump_info import DumpInfo  # noqa: F401
from .base.distribute_info import DistributeInfo  # noqa: F401
from .base.property import Property  # noqa: F401
from .base.property_type import PropertyType  # noqa: F401
from .base.name_converter import NameConverter, NamingConventionType  # noqa: F401
from .base.plugin_manager import Plugin, PluginType  # noqa: F401



from dataclasses import dataclass
from typing import List
from .meta_data_settings import MetaDataSettings


_DEFAULT_INDENT = 4


@dataclass(kw_only=True)  # https://medium.com/@aniscampos/python-dataclass-inheritance-finally-686eaf60fbb5
class ConfigurationBase:
    """
    Serves as the base for several configuration classes.
    """
    indent: int = _DEFAULT_INDENT
    """
    Whitespace indent before each property, defaults to _DEFAULT_INDENT
    """
    transformers: List[str] = None
    """
    Python scripts which can transform the provided value. The script has access to the following variables.
    - name: Property name.
    - value: Property value.
    - type: Property type value (see values PropertyType).
    - properties: List of all properties (must not be modified).

    To reflect changes to the outside of the script, the value variable must be modified.
    """
    meta_data_settings: MetaDataSettings = None
    """
    Defines which meta data to include in the generated file as comment.
    """



from typing import List, Type

from .configuration_base import _DEFAULT_INDENT, ConfigurationBase
from .language_config_naming_conventions import LanguageConfigNamingConventions
from .generator_base import GeneratorBase
from .distributor_base import DistributorBase
from .generator_configuration import GeneratorConfiguration
from .meta_data_settings import MetaDataSettings


class NoConfigNameProvidedException(Exception):
    def __init__(self):
        super().__init__('No config name has been provided')


class LanguageConfigConfiguration(ConfigurationBase):
    """
    Encapsulates the configuration properties used by the LanguageConfigBase class.
    """
    config_name: str
    """
    Name of the generated type and config. HINT: This acts more like a template
    for the type name than the real name as some conventions must be met and
    therefore the default convention specified by the deriving class of
    GeneratorBase will be used if no naming convention for the type name
    was provided (see GeneratorBase._default_type_naming_convention).
    """
    file_extension: str
    """
    Which file extension to use for the output file.
    """
    generator_type: Type[GeneratorBase]
    """
    Which generator to use to generate the config.
    """
    naming_conventions: LanguageConfigNamingConventions
    """
    Specifies which case convention to use for the properties. If not provided,
    the name as specified will be used.
    """

    distributors: List[DistributorBase]
    """
    Specifies which distributors to use for spreading the generated file.
    """

    def __init__(
        self,
        config_name: str,
        file_extension: str,
        generator_type: Type[GeneratorBase],
        indent: int=_DEFAULT_INDENT,
        transformers: List[str]=None,
        naming_conventions: LanguageConfigNamingConventions=None,
        distributors: List[DistributorBase]=None,
        meta_data_settings: MetaDataSettings=None
    ) -> None:
        super().__init__()

        self.config_name = config_name
        self.file_extension = file_extension.lstrip('.')
        self.generator_type = generator_type
        self.indent = indent
        self.transformers = transformers
        self.naming_conventions = naming_conventions
        self.distributors = distributors
        self.meta_data_settings = meta_data_settings

    def validate(self):
        """
        Validates the current configuration.

        :raises NoConfigNameProvidedException: Raised if no config name has been provided.
        """
        if not self.config_name:
            raise NoConfigNameProvidedException()
        
        # Make sure that the naming conventions are available.
        if not self.naming_conventions:
            self.naming_conventions = LanguageConfigNamingConventions()

    def get_generator_config(self) -> GeneratorConfiguration:
        """
        Creates the corresponding GeneratorConfig from the current LanguageConfigConfiguration.

        :return: GeneratorConfiguration based on the current LanguageConfigConfiguration.
        :rtype:  GeneratorConfiguration
        """
        return GeneratorConfiguration(
            indent=self.indent,
            transformers=self.transformers,
            type_name=self.config_name,
            naming_conventions=self.naming_conventions,
            meta_data_settings=self.meta_data_settings
        )



from dataclasses import dataclass


@dataclass(kw_only=True)  # https://stackoverflow.com/a/70259423
class MetaDataSettings:
    user: bool = False
    date: bool = False
    time: bool = False
    version: bool = False
    link: bool = False



class ConfigFileInfo:
    """
    Container for basic config file information.
    """
    def __init__(self, file_name: str, file_extension: str):
        """
        Constructor

        :param file_name:      File name (e.g., test-config).
        :type file_name:       str
        :param file_extension: File extension (e.g., java).
        :type file_extension:  str
        """
        self.file_name = file_name
        self.file_extension = file_extension.lstrip('.')
        self.file_name_full = f'{self.file_name}.{self.file_extension}'



import re


class NoPackageNameException(Exception):
    def __init__(self):
        super().__init__('No package name provided')


class EmptyPackageNameException(Exception):
    def __init__(self):
        super().__init__('Package name is empty')


class InvalidPackageNameException(Exception):
    def __init__(self, package_name: str, hint: str):
        message = f'Package name {package_name} is not a valid name'

        if hint:
            message = f'{message}. HINT: {hint}'
        super().__init__(message)


def evaluate_package(package_regex: str, hint: str, **props) -> str:
    """
    Checks if props contains a value for 'package' and if the package name conforms
    to the specified RegEx.

    :param package_regex: Package name RegEx to check.
    :type package_regex:  str
    :param hint:          Hint to throw if the package name doesn't conform.
    :type hint:           str

    :raises NoPackageNameException:      Raised if no package name was found.
    :raises EmptyPackageNameException:   Raised if the package name is empty.
    :raises InvalidPackageNameException: Raised if the package name does not conform to the RegEx.

    :return: Package name string.
    :rtype:  str
    """
    ATTRIBUTE_PACKAGE = 'package'

    if ATTRIBUTE_PACKAGE not in props:
        raise NoPackageNameException()
    else:
        package = props[ATTRIBUTE_PACKAGE]
    
    if not package:
        raise EmptyPackageNameException()
    elif not re.match(package_regex, package):
        raise InvalidPackageNameException(package, hint)
    return package



from __future__ import annotations
from typing import List

from .language_config_base import LanguageConfigBase
from .config import Config
from .distributor_credentials import DistributorCredentials
from .plugin_manager import Plugin

class Orchestrator:
    def __init__(self, language_configs: List[LanguageConfigBase]):
        # Make sure the configs-list is available.
        if not language_configs:
            language_configs = []

        self.language_configs = language_configs

    def dump(self) -> List[str]:
        """
        Dumps all language configs into a list of strings.

        :return: List of config strings.
        :rtype:  List[str]
        """
        return [config.dump() for config in self.language_configs]
    
    def write(self, path: str = ''):
        """
        Writes all language configs to the specified output path.

        :param path: Path to write the configs to (the directory must exist), defaults to ''
        :type path:  str, optional

        :return: The current Orchestrator instance.
        :rtype:  Orchestrator
        """
        [config.write(path) for config in self.language_configs]
        return self
    
    def distribute(self):
        """
        Distributes all generated config files via their specified distributors.

        :return: The current Orchestrator instance.
        :rtype:  Orchestrator
        """
        [config.distribute() for config in self.language_configs]
        return self

    @staticmethod
    def read_config(
        path: str,
        distributor_credentials: List[DistributorCredentials]=None,
        plugins: List[Plugin]=None,
    ):
        """
        Reads the provided YAML configuration file and generates a list of language configurations.

        :param path: Path to load the YAML file from (see example/test-config.yaml for configuration details).
        :type path:  str

        :return: Orchestrator instance.
        :rtype:  Orchestrator
        """
        return Orchestrator(Config.read(path, distributor_credentials, plugins))

    @staticmethod
    def parse_config(
        config: str | object,
        config_name: str,
        distributor_credentials: List[DistributorCredentials]=None,
        plugins: List[Plugin]=None,
    ):
        """
        Parses the provided YAML configuration string and generates a list of language configurations. 

        :param config:      YAML configuration string (see example/test-config.yaml for configuration details).
        :type config:       str
        :param config_name: Name of the generated type and config. HINT: This acts more like a template for the
                            type name than the real name as some conventions must be met and therefore the default
                            convention specified by the deriving class of GeneratorBase will be used if no naming
                            convention for the type name was provided (see
                            GeneratorBase._default_type_naming_convention).
        :type config_name:  str

        :return: Orchestrator instance.
        :rtype:  Orchestrator
        """
        return Orchestrator(Config.parse(config, config_name, distributor_credentials, plugins))



from dataclasses import dataclass
from .distributor_credentials import DistributorCredentials


@dataclass  # https://stackoverflow.com/a/70259423
class DistributeInfo:
    file_name: str
    data: str
    credentials: DistributorCredentials



from __future__ import annotations
from abc import ABC, abstractmethod
import re
from typing import List, Type

from .configuration_base import _DEFAULT_INDENT
from .generator_base import GeneratorBase
from .distributor_base import DistributorBase
from .language_config_configuration import LanguageConfigConfiguration
from .language_config_naming_conventions import LanguageConfigNamingConventions
from .name_converter import NamingConventionType
from .config_file_info import ConfigFileInfo
from .name_converter import NameConverter
from .property import Property
from .meta_data_settings import MetaDataSettings


class InvalidFileNameException(Exception):
    def __init__(self, file_name: str, pattern: str, class_type: type):
        type_name = class_type.__name__

        super().__init__(
            f'The file name "{file_name}" does not conform to the validation pattern "{pattern}" of {type_name} '
            'as it might be problematic to import the module later on. The output-filename convention can be '
            'specified using the "file_naming" property.'
        )


class LanguageConfigBase(ABC):
    """
    Abstract class which serves as the base for all language specific config classes. The LanguageConfigBase holds all
    required information (language type, naming convention, generator, ...) to generate a config file.
    """

    def __init__(
        self,
        config_name: str,
        properties: List[Property],
        indent: int = _DEFAULT_INDENT,
        transformers: List[str] = None,
        naming_conventions: LanguageConfigNamingConventions = None,
        distributors: List[DistributorBase] = None,
        meta_data_settings: MetaDataSettings = None,
        additional_props = {},
    ):
        """
        Constructor

        :param config_name:        Language config name. The config name acts more like a template as it might
                                   be changed by the naming convention rules or the specific language config
                                   implementation (e.g. test-config might become e.g. TestConfig).
        :type config_name:         str
        :param properties:         List of properties.
        :type properties:          List[Property]
        :param indent:             Property indent for the generated config, defaults to _DEFAULT_INDENT
        :type indent:              int, optional
        :param transformers:       Python functions which can transform the provided value, defaults to None
        :type transformers:        List[str], optional
        :param naming_conventions: Naming convention to use for the generated config file, defaults to None
        :type naming_conventions:  LanguageConfigNamingConventions, optional
        :param distributors:       List of distributors, defaults to None
        :type distributors:        List[DistributorBase], optional
        :param additional_props:   All props that might by needed by the derivating class, defaults to {}
        :type additional_props:    dict, optional
        """
        config = LanguageConfigConfiguration(
            config_name=config_name,
            file_extension=self._file_extension(),
            generator_type=self._generator_type(),
            indent=indent,
            transformers=transformers,
            naming_conventions=naming_conventions,
            distributors=distributors,
            meta_data_settings=meta_data_settings,
        )

        # Make sure, config is valid.
        config.validate()

        self.generator = config.generator_type(
            config.get_generator_config(),
            properties,
            additional_props,
        )
        file_naming_convention = config.naming_conventions.file_naming_convention
        file_naming_convention = file_naming_convention \
            if file_naming_convention else self._default_file_naming_convention()

        self.config_info = ConfigFileInfo(
            # Convert config file name according to naming convention if a convention was provided. Otherwise, just use
            # the config name directly.
            NameConverter.convert(
                config.config_name,
                file_naming_convention if file_naming_convention else NamingConventionType.SNAKE_CASE,
            ),
            config.file_extension,
        )
        self.distributors = distributors if distributors else []

        # Check output file naming.
        self._check_file_name()

    def dump(self) -> str:
        """
        Generates a config file string.

        :return: Config file string.
        :rtype:  str
        """
        return self.generator.dump()
    
    def write(self, path: str = ''):
        """
        Generates a config file string and writes the config file to the provided directory.

        :param path: Directory to write the file to, defaults to ''
        :type path:  str, optional

        :return:     The current LanguageConfigBase instance.
        :rtype:      LanguageConfigBase
        """
        path = path.rstrip('/').rstrip('\\')  # Strip right-side slashes.
        path = f'{path}/{self.config_info.file_name_full}'

        with open(path, 'w') as f:
            f.write(self.dump())
        return self
    
    def distribute(self):
        """
        Distributes the generated config file via the specified distributors.

        :return: The current LanguageConfigBase instance.
        :rtype:  LanguageConfigBase
        """
        data = self.dump()

        [distributor.distribute(self.config_info.file_name_full, data) for distributor in self.distributors]            
        return self

    @abstractmethod
    def _file_extension(self) -> str:
        pass

    @abstractmethod
    def _generator_type(self) -> Type[GeneratorBase]:
        pass

    @abstractmethod
    def _default_file_naming_convention(self) -> NamingConventionType:
        pass

    @abstractmethod
    def _allowed_file_name_pattern(self) -> str:
        """
        Abstract method which must be implemented by the deriving class to provide a RegEx string which describes which
        file name patterns are allowed for the output file name (without extension).

        :return: Allowed file name pattern.
        :rtype:  str
        """
        pass

    def _check_file_name(self) -> None:
        """
        Checks if the config file name matches the pattern defined by the deriving class.

        :raises InvalidFileNameException: Thrown if the file name is not valid.
        """
        pattern = self._allowed_file_name_pattern()

        if not re.match(pattern, self.config_info.file_name):
            raise InvalidFileNameException(self.config_info.file_name, pattern, type(self))



from pathlib import Path
from setuptools import setup, find_packages

# Setup process taken from here: https://www.freecodecamp.org/news/build-your-first-python-package/.

DESCRIPTION = 'ninja-bear keeps your language specific configs in sync'
LONG_DESCRIPTION = Path(__file__).parent.absolute().joinpath('README.md').read_text()

# Get version.
try:
    with open('src/ninja_bear/base/info.py') as fp:
        info = {}
        exec(fp.read(), info)
        VERSION = info['VERSION']
except Exception as e:
    print(e)
    exit(-1)

setup(
    name='ninja-bear', 
    version=VERSION,
    author='monstermichl',
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    package_dir={'': 'src'},
    packages=find_packages(where='src'),
    py_modules=['ninja_bear'],
    entry_points = {
        'console_scripts': ['ninja-bear=ninja_bear.cli:main'],
    },
    install_requires=[
        'pyyaml >= 6.0.1',
        'schema >= 0.7.5',
        'importlib-metadata >= 8.5.0',
    ],
    extras_require={
        'dev': [
            'wheel>=0.41.1',
            'twine>=4.0.2',
            'ruff>=0.0.47',
            'coverage>=7.2.7',
        ],
    },
    python_requires='>=3.10',
    classifiers= [
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    url = 'https://github.com/monstermichl/ninja-bear.git',
    keywords = [
        'ninja-bear',
        'multi',
        'language',
        'generator',
        'distributor',
        'configuration',
    ],
)



from dataclasses import dataclass
from typing import Dict, List
from .property import Property


@dataclass  # https://stackoverflow.com/a/70259423
class DumpInfo:
    type_name: str
    properties: List[Property]
    indent: int
    additional_props: Dict
